# Based on https://machinelearningmastery.com/how-to-develop-a-conditional-generative-adversarial-network-from-scratch/

# The Idea for this is to embed the latent-space input coordinates into the random noise and the real dataset images so that the generator works with them as well as the discriminator knowing about it.

import numpy as np
import tensorflow.keras as keras
import tensorflow as tf
import os
import random
import bitstring

from tensorflow.python.framework.ops import Tensor
from tensorflow.python.keras.engine.training import Model

fileDir = os.path.dirname(os.path.abspath(__file__))

# define standalone discriminator
def define_discriminator(in_shape=(12, 32, 1)):
    # weight initialization
    init = keras.initializers.RandomNormal(stddev=0.02)
    # define model
    model = keras.Sequential()
    # downsample to 3x8
    model.add(keras.layers.Conv2D(64, (4, 4), strides=(2, 2),
              padding='same', kernel_initializer=init, input_shape=in_shape))
    model.add(keras.layers.BatchNormalization())
    model.add(keras.layers.LeakyReLU(alpha=0.2))
    # classifier
    model.add(keras.layers.Flatten())
    model.add(keras.layers.Dense(58, activation='relu'))
    model.add(keras.layers.Dense(1, activation='sigmoid'))
    # compile model
    opt = keras.optimizers.Adam(lr=0.0002, beta_1=0.5)
    model.compile(loss='binary_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model

# define the standalone generator model
def define_generator(latent_dim):
    inputLayer = keras.layers.Input((4, 32, 1), name="gen_Input_Reshape")
    x = keras.layers.Reshape((4, 32, 1), name="gen_Reshape_1")(inputLayer)
    x = keras.layers.LeakyReLU(alpha=0.2, name="gen_LeakyReLu_1")(x)
    # upsample to 8x32
    x = keras.layers.Conv2DTranspose(
        256, (4, 4), strides=(2, 1), padding='same', name="gen_Conv2DTranspose_1")(x)
    x = keras.layers.LeakyReLU(alpha=0.2, name="gen_LeakyRelu_2")(x)

    x = keras.layers.LeakyReLU(alpha=0.2, name="gen_LeakyReLu_3")(x)

    x = keras.layers.Conv2D(
        128, (5,5), activation='tanh', padding='same', name="gen_Conv2D_1")(x)
    x=keras.layers.LeakyReLU(alpha=0.2, name="gen_LeakyRelu_4")(x)
    x=keras.layers.Dense(1,activation='tanh')(x)
    concatenator = keras.layers.concatenate(
        [x, inputLayer], axis=1)
    model = Model(inputs=[inputLayer], outputs=[concatenator],name="Generator")
    print(model.summary())
    return model

# define the combined generator and discriminator model, for updating the generator
def define_gan(generator, discriminator):
    # make weights in the discriminator not trainable
    discriminator.trainable = False
    # connect them
    model = keras.Sequential(name="GAN")
    # add generator
    model.add(generator)
    # add the discriminator
    model.add(discriminator)
    # compile model
    opt = keras.optimizers.Adam(lr=0.0002, beta_1=0.5)
    model.compile(loss='binary_crossentropy', optimizer=opt)
    print(model.summary())
    return model

# get input data from file
def loadDataFromFile():
    # Load data from testfile
    dataFile = open(os.path.join(fileDir, "testdata.txt"), "r")
    dataString = dataFile.read()
    lines = dataString.split("\n")
    patterns = []
    for line in lines:
        if(line != ""):
            pattern = []
            allBits = line.split(",")[:-1]
            for rownumber in range(12):
                row = []
                for stepnumber in range(32):
                    row.append(allBits[rownumber*32+stepnumber])
                pattern.append(row)
            patterns.append(pattern)
    return patterns

# transform input data
def load_real_samples(input_data):
    # expand to 3d, e.g. add channels
    X = keras.backend.expand_dims(input_data, axis=-1)
    # convert from ints to floats
    X = X.numpy().astype('float32')
    # scale from [0,1] to [-1,1]
    X = (X - 0.5) * 2
    print("Loaded Real Data:")
    print(tf.shape(X))
    return X

# select real samples
def generate_real_samples(dataset, n_samples):
    # choose random instances
    ix = np.random.randint(0, dataset.shape[0], n_samples)
    # select images
    X = dataset[ix]
    # generate class labels
    y = keras.backend.ones((n_samples, 1))
    print("Generating real samples. First sample from data:")
    return X, y

def getRandomLatentCoords(number):
    result = []
    for i in range(number):
        coords = []
        for x in range(4):
            coords.append(convertf32ToBools(random.uniform(-8.0, 8.0)))
        result.append(coords)
    return result

# Converts the given float to its bitwise representation and returns it as an array.
def convertf32ToBools(num):
    result = []
    b = bitstring.pack('>f', num)
    blist = list(b)
    for char in blist:
        result.append(float(char))
    if (len(result) != 32):
        print("something went wrong converting f32 to bitarray")
    return result

# use the generator to generate n fake examples, with class labels
def generate_fake_samples(generator, latent_dim, n_samples, input_data):
    # generate points in latent space
    x_input = getRandomLatentCoords(n_samples)
    # predict outputs
    X = generator.predict(tf.convert_to_tensor(x_input))
    # print("Prediction stuff: ")
    # print(tf.shape(X))
    # print(X)
    # create class labels
    y = keras.backend.zeros((n_samples, 1))
    return X, y


# prints an output tensor from the generator
def print_Prediction(tensor: Tensor):
    for pattern in tensor:
        print("{")
        for row in pattern:
            rowstr = ""
            for step in row:
                if(step > 0):
                    rowstr = rowstr+"1,"
                else:
                    rowstr = rowstr+"0,"
            print(rowstr)
        print("}")
        
# train the generator and discriminator
def train(g_model, d_model, gan_model, dataset, latent_dim, input_data, n_epochs=50, n_batch=256):
    bat_per_epo = int(dataset.shape[0] / n_batch)
    half_batch = int(n_batch / 2)
    # manually enumerate epochs
    for i in range(n_epochs):
        # enumerate batches over the training set
        for j in range(bat_per_epo):
            # get randomly selected 'real' samples
            X_real, y_real = generate_real_samples(dataset, half_batch)
            print("Real Sample(" + str(y_real[0]) + "):")
            print_Prediction([X_real[0]])
            # update discriminator model weights
            d_loss1, _ = d_model.train_on_batch(X_real, y_real)
            # generate 'fake' examples
            X_fake, y_fake = generate_fake_samples(
                g_model, latent_dim, half_batch, input_data)
            # update discriminator model weights
            #print("Fake Sample(" + str(y_fake[0]) + "):")
            #print_Prediction([X_fake[0]])
            d_loss2, _ = d_model.train_on_batch(X_fake, y_fake)
            # prepare points in latent space as input for the generator
            X_gan = getRandomLatentCoords(n_batch)
            # create inverted labels for the fake samples
            y_gan = keras.backend.ones((n_batch, 1))
            # update the generator via the discriminator's error
            # print("Feeding batch into gan: ")
            # print("X: %s, Y: %s", tf.shape(X_gan), tf.shape(y_gan))
            g_loss = gan_model.train_on_batch(
                tf.convert_to_tensor(X_gan), y_gan)
            # summarize loss on this batch
            print('>%d, %d/%d, d1=%.3f, d2=%.3f g=%.3f' %
                  (i+1, j+1, bat_per_epo, d_loss1, d_loss2, g_loss))
        pred_input = getRandomLatentCoords(1)
        print("Random Prediction Input:")
        print_Prediction(pred_input)
        print("Random Prediction:")
        print_Prediction(g_model.predict(
            (pred_input)))

    # save the generator model
    converter = tf.lite.TFLiteConverter.from_keras_model(g_model)
    tflite_model = converter.convert()
    open("8_Bools_modded_GAN_Trainer.tflite", "wb").write(tflite_model)


# size of the latent space
latent_dim = 4*32
input_data = loadDataFromFile()
# create the discriminator
print("Defining discriminator")
discriminator = define_discriminator()
# create the generator
print("Defining generator")
generator = define_generator(latent_dim)
# create the gan
print("Defining GAN")
gan_model = define_gan(generator, discriminator)
# load image data
print("Loading Data")
dataset = load_real_samples(input_data)
# train model
print("Starting Training")
train(generator, discriminator, gan_model, dataset, latent_dim, input_data)
