/* Copyright 2020 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#include "main_functions.h"

#include "constants.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "model.h"
#include "output_handler.h"
#include "tensorflow/lite/micro/all_ops_resolver.h"
#include "tensorflow/lite/micro/micro_error_reporter.h"
#include "tensorflow/lite/micro/micro_interpreter.h"
#include "tensorflow/lite/schema/schema_generated.h"
#include "tensorflow/lite/version.h"

// Globals, used for compatibility with Arduino-style sketches.
namespace {
tflite::ErrorReporter* error_reporter = nullptr;
const tflite::Model* model = nullptr;
tflite::MicroInterpreter* interpreter = nullptr;
TfLiteTensor* input = nullptr;
TfLiteTensor* output = nullptr;

// Create an area of memory to use for input, output, and intermediate arrays.
//  After allocating tensors
// you can retrieve this value by invoking interpreter.arena_used_bytes().
const int kModelArenaSize = 7180;
// Extra headroom for model + alignment + future interpreter changes.
const int kExtraArenaSize = 2048;
const int kTensorArenaSize = kModelArenaSize + kExtraArenaSize;

uint8_t tensor_arena[kTensorArenaSize];
}  // namespace

// The name of this function is important for Arduino compatibility.
void setup() {
  // Set up logging. Google style is to avoid globals or statics because of
  // lifetime uncertainty, but since this has a trivial destructor it's okay.
  // NOLINTNEXTLINE(runtime-global-variables)
  static tflite::MicroErrorReporter micro_error_reporter;

  error_reporter = &micro_error_reporter;
  // Map the model into a usable data structure. This doesn't involve any
  // copying or parsing, it's a very lightweight operation.
  model = tflite::GetModel(g_model);
  if (model->version() != TFLITE_SCHEMA_VERSION) {
    TF_LITE_REPORT_ERROR(error_reporter,
                         "Model provided is schema version %d not equal "
                         "to supported version %d.",
                         model->version(), TFLITE_SCHEMA_VERSION);
    return;
  }

  // This pulls in all the operation implementations we need.
  // NOLINTNEXTLINE(runtime-global-variables)
  static tflite::AllOpsResolver resolver;

  // Build an interpreter to run the model with.
  static tflite::MicroInterpreter static_interpreter(
      model, resolver, tensor_arena, kTensorArenaSize, error_reporter);
  interpreter = &static_interpreter;

  // Allocate memory from the tensor_arena for the model's tensors.
  TfLiteStatus allocate_status = interpreter->AllocateTensors();
  if (allocate_status != kTfLiteOk) {
    TF_LITE_REPORT_ERROR(error_reporter, "AllocateTensors() failed");
    return;
  }
  printf("Tensor Allocation Status: %d\n", allocate_status);
  // Obtain pointers to the model's input and output tensors.
  input = interpreter->input(0);
  output = interpreter->output(0);
  printf(
      "Initialized model with a input size of %d and a output size of %d.\n\n",
      interpreter->outputs_size(), interpreter->inputs_size());
  printf("Model version: %d\n", model->version());
  printf("arena size: %d\n", interpreter->arena_used_bytes());
  printf("Checking input and output tensor properties:\n");
  // Make sure the input has the properties we expect
  TFLITE_CHECK_NE(nullptr, input);
  // The property "dims" tells us the tensor's shape. It has one element for
  // each dimension. Our input is a 2D tensor containing 1 element, so "dims"
  // should have size 2.
  printf("Input dimension size: %d\n", input->dims->size);
  printf("Input data size: %d\n", input->dims->data[1]);
  printf("Input byte size: %d\n", input->bytes);
  printf("Input type: %d\n", input->type);

  printf("Output dimension size: %d\n", output->dims->size);
  printf("Output data size: %d\n", output->dims->data[1]);
  printf("Output byte size: %d\n", output->bytes);
  printf("Output type: %d\n", output->type);

  printf("tensors Verified.\n");
}

float getRandomCoord(float maxCoord, float minCoord, int seed) {
  srand(seed);
  float coord =(float)rand()/(float)(RAND_MAX);
  printf("%lf\n",coord);
  coord = coord * (maxCoord - minCoord) + minCoord;
  printf("%lf\n",coord);
  return coord;
}

// The name of this function is important for Arduino compatibility.
void loop() {
  unsigned long startTime = xTaskGetTickCount();
  // Place our calculated x value in the model's input tensor
  int seed = time(NULL);
  input->data.f[0] = getRandomCoord(0.0, 1250, seed);
  input->data.f[1] = getRandomCoord(0.0, 950, seed + 111);
  printf("Coords: %lf, %lf\n", input->data.f[0], input->data.f[1]);

  // Run inference, and report any error
  TfLiteStatus invoke_status = interpreter->Invoke();
  if (invoke_status != kTfLiteOk) {
    // TF_LITE_REPORT_ERROR(error_reporter,
    //                     "Invoke failed on input->data.f: %f,%f,%f,%f\n",
    //                     static_cast<double>(input->data.f[0]),
    //                     static_cast<double>(input->data.f[1]);
    return;
  }

  // print the results in integer form
  printf("used space: %d\n", interpreter->arena_used_bytes());
  printf("in %lu ticks(%lu - %lu)\n", xTaskGetTickCount() - startTime,
         (unsigned long)xTaskGetTickCount(), startTime);
  printf("output at address: %p\n", output->data.f);

  // print the resulting Matrix:
  for (int row = 0; row < 8; row++) {
    char str[64] = "";
    for (int col = 0; col < 32; col++) {
      if ((output->data.f[col + row * 32]) >= 0.5) {
        str[col * 2] = '1';
        str[col * 2 + 1] = ',';
      } else {
        str[col * 2] = '0';
        str[col * 2 + 1] = ',';
      }
    }
    printf("%s\n", str);
  }
  vTaskDelay(1000 / portTICK_PERIOD_MS);
}
