# based on https://towardsdatascience.com/drum-patterns-from-latent-space-23d59dd9d827
import tensorflow.keras as keras
import tensorflow as tf
import os
import numpy as np
import random
import matplotlib.pyplot as plt
import math
import re
import bitstring
from tensorflow.python.keras.backend import dtype

# directory of this script
fileDir = os.path.dirname(os.path.abspath(__file__))


# converts a integer-colums-representation of a matrix to a matrix of floats of either 0.0 or 1.0.
def col_ints_to_matrix(colArr):
    mat = []
    for i in range(8):
        mat.append([])
    for colInt in colArr:
        binarr = convertf32ToBools(float(colInt))
        for i in range(8):
            mat[i].append(binarr[i])
    return mat


# Converts the given float to its bitwise representation and returns it as an array.
def convertf32ToBools(num):
    result = []
    b = bitstring.pack('>f', num)
    blist = list(b)
    for char in blist:
        result.append(float(char))
    if (len(result) != 32):
        print("something went wrong converting f32 to bitarray")
    return result


# prints a textual representation of the data
def print_pattern(pattern):
    if(len(pattern) == 256):
        for row in range(8):
            rowstr = ""
            for stepnr in range(32):
                step = pattern[row*32+stepnr]
                if(step >= 0.5):
                    rowstr = rowstr + "1,"
                else:
                    rowstr = rowstr + "0,"
            print(rowstr)
    else:
        if(isinstance(pattern[0], np.float32) or (isinstance(pattern[0], np.int32))):
            pattern = col_ints_to_matrix(pattern)
        for row in pattern:
            rowstr = ""
            for step in row:
                if(step >= 0.5):
                    rowstr = rowstr + "1,"
                else:
                    rowstr = rowstr + "0,"
            print(rowstr)
    print("\n")


# generate the full model
def generate_model() -> keras.Model:
    model = keras.Sequential([
        keras.Input(shape=(256,)),
        keras.layers.Dense(512, activation="relu", name="enc_dense_input", dtype=tf.dtypes.float32),
        keras.layers.Dense(256, activation="relu", name="enc_dense_1", dtype=tf.dtypes.float32),
        keras.layers.Dense(128, activation="relu", name="enc_dense_2", dtype=tf.dtypes.float32),
        keras.layers.Dense(64, activation="relu", name="enc_dense_3", dtype=tf.dtypes.float32),
        keras.layers.Dense(2, activation="relu", name="enc_output"),
        keras.layers.Dense(64, activation="relu", name="dec_dense_1"),
        keras.layers.Dense(128, activation="relu", name="dec_dense_2"),
        keras.layers.Dense(512, activation="relu", name="dec_dense_3"),
        keras.layers.Dense(256, activation="sigmoid", name="dec_output_dense")
    ], name="Full_Autoencoder")
    model.compile(optimizer="adam", metrics=["acc"], loss="mean_squared_error")
    return model

# Converts the given integer to its boolean representation and returns it as an array.
def convertToBools(colInt):
    row = []
    for i in range(32):
        rest = colInt % 2
        row.append(float(rest))
        colInt = float(math.trunc(colInt/2))
    return row

# Load data from testfile
def loadDataFromFile():
    # Load data from testfile
    dataFile = open(os.path.join(fileDir, "testdata.txt"), "r")
    dataString = dataFile.read()
    lines = dataString.split("\n")
    data = {}
    data["patterns"] = []
    for line in lines:
        allfields = re.split('\t|,|\n| ', line)
        pattern = np.zeros((256))
        for col in range(32):
            rowBools = convertToBools(float(allfields[col]))
            for nr in range(8):
                pattern[nr*32 + col] = rowBools[nr]
        data["patterns"].append(np.array(pattern).astype(np.float32))
    data["patterns"] = tf.convert_to_tensor(data["patterns"])
    return data

# Saves the given model as tflite model with the given name.
def save_model_as_tflite(model, name):
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    tflite_model = converter.convert()
    open(os.path.join(fileDir, "tflite_models/" +
                      name + ".tflite"), "wb").write(tflite_model)


model = generate_model()
data = loadDataFromFile()
print(model.summary())

print("fitting model to data")
history = model.fit(x=data["patterns"],
                    y=data["patterns"], epochs=1000, verbose=2)
print("model fitting finished")


# plot the history
# accuracy
plt.plot(history.history['acc'])
plt.title('model acc')
plt.ylabel('acc')
plt.xlabel('epoch')
plt.legend(['train', 'test'], loc='upper left')
plt.show()

# loss
plt.plot(history.history['loss'])
plt.title('model loss')
plt.ylabel('loss')
plt.xlabel('epoch')
plt.legend(['train', 'test'], loc='upper left')
plt.show()

# get the seperate encoder and decoder:
enc = keras.Model(inputs=model.input,
                  outputs=model.layers[4].output, name="encoder")
dec_input = keras.Input([None, 1], tensor=model.layers[5].input)
dec = keras.Model(inputs=dec_input,
                  outputs=model.layers[8].output, name="decoder")
print(enc.summary())
print(dec.summary())

# test the results
print("Decoder-test:")
print_pattern(np.array(dec.predict(
    [[random.uniform(-8.0, 8.0), random.uniform(-8.0, 8.0)]])[0]))
print("encoder-test:")
testpattern = data['patterns'][random.randint(0, len(data['patterns']))]
print("input pattern: ")
print_pattern(np.array(testpattern))
inp = tf.convert_to_tensor([testpattern])
print("prediction:")
print((enc.predict(np.array(inp)))[0])
print("full test:")
testpattern_2 = tf.convert_to_tensor(
    [data['patterns'][random.randint(0, len(data['patterns']))]])
print("input-pattern: ")
print_pattern(np.array(testpattern_2)[0])
print("output-pattern:")
print_pattern(np.array((model.predict(testpattern_2))[0]))

print("Analysing models")

# plot 100 points: coords->enc->decoder->coords
analyzing_coords = []
for i in range(100):
    analyzing_coords.append(
        [[random.uniform(-8.0, 8.0)], [random.uniform(-8.0, 8.0)]])
resulting_coords = (np.array(enc.predict(
    dec.predict(analyzing_coords)))-4.0)*2.0
average_miss = 0.0
for ind in range(len(analyzing_coords)):
    diff_x = resulting_coords[ind][0]-analyzing_coords[ind][0]
    diff_y = resulting_coords[ind][1]-analyzing_coords[ind][1]
    average_miss = average_miss + math.sqrt((diff_x)**2+(diff_y)**2)
average_miss = average_miss/100
print("average miss:")
print(average_miss)

for i in range(len(analyzing_coords)):
    x = [analyzing_coords[i][0][0], resulting_coords[i][0]]
    y = [analyzing_coords[i][1][0], resulting_coords[i][1]]
    plt.plot(x, y, 'b')
    plt.plot(analyzing_coords[i][0], analyzing_coords[i][1], 'go')
    plt.plot(resulting_coords[i][0], resulting_coords[i][1], 'ro')

plt.plot([-8.0, -4.0, 0, 4, 8], [-8.0, -4.0, 0, 4, 8], 'yo')

plt.title('coords- before and after')
plt.ylabel('y')
plt.xlabel('x')
plt.show()

# generate some patterns from patterns and see the result change
for i in range(5):
    selectedPattern = tf.convert_to_tensor(
        [data['patterns'][random.randint(0, len(data['patterns']))]])
    resultpattern = model.predict([selectedPattern])
    print("round ", i, " results: ")
    print("input:")
    print_pattern(np.array(selectedPattern[0]))
    print("result:")
    print_pattern(np.array(resultpattern[0]))

# save tflite models for encoder and decoder:
save_model_as_tflite(enc, "2D_encoder_smaller_kernels_more_layers")
save_model_as_tflite(dec, "2D_decoder_smaller_kernels_more_layers")
