import numpy as np
import tensorflow as tf
import playsound
import time
import threading
import os

import sounddevice as sd
from scipy.io.wavfile import write
from pathlib import Path
import random
import math

from tensorflow.lite.python import convert

fs = 44100  # Sample rate
seconds = 2.0  # Duration of recording in seconds - includes .2 seconds buffer to fully play sounds last in queue

stepCount = 32
instrumentCount = 8

# input coord space
latent_space = [[-8.0, 8.0], [-8.0, 8.0], [-8.0, 8.0], [-8.0, 8.0]]

# This sets the audiomix as input to record. Will propably be different on other machines.
sd.default.device = [2, 4]

# absolute path to search for models
model_search_path = ""

# list of sounds to be played by this script
sounds = ["instruments/1.mp3",
          "instruments/2.mp3",
          "instruments/3.mp3",
          "instruments/4.mp3",
          "instruments/5.mp3",
          "instruments/6.mp3",
          "instruments/7.mp3",
          "instruments/8.mp3"]

# all models (including decoders)
all_models_found = list(
    Path(model_search_path).rglob("*.[tT][fF][lL][iI][tT][eE]"))

# filtering out decoders
models_found = []
for modelpath in all_models_found:
    if not "encoder" in modelpath.stem and not "initial_test" in modelpath.stem:
        models_found.append(modelpath)

# Print a text-representation of the matrix
def printMatrix(matrix):
    output = ""
    for row in matrix:
        for step in row:
            output = output + str(step) + ","
        output = output + "\n"
    print(output)

# Generate random coordinates in an array with the given dimensions.
# Coordinates will range in the latent space dimensions given by latent_space array.
def getRandomCoords(dims):
    result = []
    for dim in range(dims):
        result.append(random.uniform(
            latent_space[dim][0], latent_space[dim][1]))
    return np.array([result]).astype(np.float32)

# Checks if the given pattern is all zeroes.
def isEmpty(pattern):
    res = 0
    for row in pattern:
        for step in row:
            res = res + step
    return res == 0

# Converts the given Integer to an array containing its boolean representation. 
# Converts the given integer to its boolean representation and returns it as an array.
def convertToBools(colInt):
    row = []
    for i in range(32):
        rest = colInt % 2
        row.append(rest)
        colInt = math.trunc(colInt/2)
    return row

# Generate and format pattern into 8x32 matrix
def generatePattern(input_data):
    interpreter.set_tensor(input_details[0]['index'], input_data)
    interpreter.invoke()
    output_data = np.array(((interpreter.get_tensor(
        output_details[0]['index'])) + 1)/2).astype(int)
    if(output_data.size == 448):
        output_data = output_data[0][0: 256]
    elif(output_data.size == 32):
        temp = np.zeros((8, 32))
        for col in range(32):
            colBools = convertToBools(output_data[0][col])
            for row in range(8):
                temp[row][col] = colBools[row]
        output_data = temp
    elif(output_data.size == 8):
        temp = np.zeros((8, 32))
        for row in range(8):
            rowBools = convertToBools(output_data[0][row])
            for col in range(32):
                temp[row][col] = rowBools[row]
        output_data = temp
    return np.reshape(output_data, [8, 32])

# Player thread class
class player (threading.Thread):
    def __init__(self, threadID, name, counter):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.name = name
        self.counter = counter
        self.batchRecorded = False
        self.recordStarted = False
        self.laps = 0
        self.pattern = []

    def run(self):
        while(True):
            if(len(self.pattern) == 0):
                time.sleep(0.1)
            else:
                for tick in range(stepCount):
                    if(tick == 0 and self.batchRecorded == False and self.recordStarted == False):
                        recorder.startRecording()
                        self.laps=1
                        self.recordStarted = True
                    for instr in range(instrumentCount):
                        if(pattern[instr][tick] >= 0.5):
                            playsound.playsound(sounds[instr], False)
                    time.sleep(0.06)
                if(self.batchRecorded == False and self.recordStarted == True and self.laps == 2):
                    recorder.stopRecording(np.array_str(coords))
                    self.batchRecorded = True
                    self.recordStarted = False
                elif(self.batchRecorded == False and self.recordStarted == True and self.laps < 2):
                    self.laps = self.laps+1

# instrument thread class
class instrumentThread(threading.Thread):
    def __init__(self, threadID, name, counter, soundPath):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.name = name
        self.counter = counter
        self.soundPath = soundPath

    def run(self):
        if self.counter == 0:
            self.counter = 1
        else:
            playsound.playsound(self.soundPath, False)

# recorder thread class
class recordThread(threading.Thread):
    def __init__(self, threadID, name):
        threading.Thread.__init__(self)
        self.active_recording = {}
        self.threadID = threadID
        self.name = name

    def startRecording(self):
        print("Starting Recording")
        self.active_recording = sd.rec(
            int(seconds * fs), samplerate=fs, channels=2, blocking=False)

    def stopRecording(self, filename):
        sd.stop()
        print("Saving Recording to file")
        write(os.path.join(recordingPath, filename + '.wav'),
              fs, self.active_recording)


# initialize instruments
instruments = []
instcounter = 0
for path in sounds:
    instcounter = instcounter + 1
    instruments.append(instrumentThread(
        instcounter, "instrument_"+path, instcounter, path))

# initialize recorder
recorder = recordThread(12121221, "Recording Thread")
recordThread.start(recorder)

# start the player
playerthread = player(1, "Player Thread", 1)
player.start(playerthread)

# go through all models in the given path
for model_path in models_found:
    full_model_path = os.path.abspath(model_path)
    print("now working with model '" + model_path.name +
          "' from '" + full_model_path + "'")
    # create subfolder for model
    recordingPath = "audio_generation/audiofiles/" + model_path.stem
    os.makedirs(recordingPath, exist_ok=True)

    interpreter = tf.lite.Interpreter(model_path=full_model_path)
    interpreter.allocate_tensors()

    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    input_shape = input_details[0]['shape']

    # generate a random sample
    input_data = getRandomCoords(input_shape[1])
    coords = input_data

    pattern = generatePattern(input_data)
    playerthread.pattern = pattern
    printMatrix(pattern)

    # continue to generate an play patterns depending on user input until 'next' is entered
    while(True):
        inputstr = input("enter coordinates to start a new pattern(format: " +
                         str(input_shape) + ") or 'next' to go to the next model:\n'")
        if(inputstr == "next"):
            break
        elif(len(inputstr) < 8):
            coords = getRandomCoords(input_shape[1])
        else:
            coords = np.array([inputstr.split(' ')]).astype(np.float32)
        print("resetting batchrecorded")
        playerthread.batchRecorded = False
        playerthread.recordStarted = False
        playerthread.laps = 0
        print("generating for coords:")
        print(coords)
        pattern = generatePattern(coords)
        if(isEmpty(pattern)):
            pattern = generatePattern(coords)
        playerthread.pattern = pattern
        printMatrix(pattern)
